@@ -83,6 +83,7 @@ devconfig(char *instr, dev_info_t *dinfo)
     dinfo->banner = NULL;
     dinfo->trace_read = NULL;
     dinfo->trace_write = NULL;
+    dinfo->trace_both = NULL;
     pos = strtok_r(str, ", \t", &strtok_data);
     while (pos != NULL) {
 	if (strcmp(pos, "300") == 0) {
@@ -151,22 +152,17 @@ devconfig(char *instr, dev_info_t *dinfo)
 	    dinfo->allow_2217 = 1;
 	} else if (strcmp(pos, "NOBREAK") == 0) {
 	    dinfo->disablebreak = 1;
-	} else if (strncmp(pos, "tr:", 3) == 0) {
+	} else if (strncmp(pos, "tr=", 3) == 0) {
 	    /* trace read, data from the port to the socket */
 	    dinfo->trace_read = find_tracefile(pos + 3);
-	    /* FIXME - error handling */
-	} else if (strncmp(pos, "tw:", 3) == 0) {
-	    /* trace read, data from the socket to the port */
+	} else if (strncmp(pos, "tw=", 3) == 0) {
+	    /* trace write, data from the socket to the port */
 	    dinfo->trace_write = find_tracefile(pos + 3);
-	    /* FIXME - error handling */
-	} else if (strncmp(pos, "tb:", 3) == 0) {
+	} else if (strncmp(pos, "tb=", 3) == 0) {
 	    /* trace both directions. */
-	    dinfo->trace_read = find_tracefile(pos + 3);
-	    dinfo->trace_write = dinfo->trace_read;
-	    /* FIXME - error handling */
+	    dinfo->trace_both = find_tracefile(pos + 3);
 	} else if ((dinfo->banner = find_banner(pos))) {
 	    /* It's a banner to display at startup, it's already set. */
-	    /* FIXME - error handling */
 	} else {
 	    rv = -1;
 	    goto out;
